
# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.

import numpy as np

from .Layer import Layer
import Models.Layouts as Layouts
from Controllers.TensorFormat import LayerIOFormat, LayerIOFormatPair
from Controllers.Tensor import UnpopulatedTensor, PopulatedTensor

class DetectionOutput(Layer):
    _supported_IO_formats = [
        (LayerIOFormat.nhwc_compact(), LayerIOFormat.nhwc_compact(),
         LayerIOFormat.nhwc_compact(), LayerIOFormat.nchw_compact())
        ]

    def __init__(self, *args):
        super().__init__(*args)

    def loadTensorSizes(self, tensorDict):
        super().loadTensorSizes(tensorDict)

        # Because Myriad cannot return a variable tensor size,
        # we force the output of this layer to be the maximum possible
        self.outputTensorSizes = []
        assert(len(self.outputTensorNames) == 1)
        for name in self.outputTensorNames:
            self.outputTensorSizes.append((1, 1, self.getMaxDetections(), 7))

    def load_parameters(self, detection_param):
        share_location = 1 if detection_param.share_location else 0

        # TODO: Find a better way to check for this field. Caffe protobuf classes surely have one
        if(hasattr(detection_param,"objectness_score")):
            # objectness_score is a specific field for the RefineDet specific detection
            # this has this extra parameter, that needs to be passed to the executor of the layer,
            # but the compiler needs to be aware that this version may accept num outputs of up to 5 
            objectness_score = detection_param.objectness_score 
            self._supported_IO_formats = [
                (LayerIOFormat.nhwc_compact(), LayerIOFormat.nhwc_compact(),
                LayerIOFormat.nhwc_compact(), LayerIOFormat.nhwc_compact(),
                LayerIOFormat.nhwc_compact(), LayerIOFormat.nchw_compact())
                ]
        else:
            #TODO: will assume 0 means no RefineDet type DetectionOutput. Can make this assumption?
            objectness_score = 0 

        det_out_dtype = np.dtype("<i4, <i4, <i4, <f4, <i4, <i4, <i4, <f4, <i4, <f4, <i4, <f4")


        # The weiliu Caffe implementation of SSD has hardoced clip = False for
        # the DetectionOutput layer.
        clip_box = 0
        self.params= np.array((detection_param.num_classes,
            share_location,
            detection_param.background_label_id,
            detection_param.nms_param.nms_threshold,
            detection_param.nms_param.top_k,
            detection_param.code_type,
            detection_param.keep_top_k,
            detection_param.confidence_threshold,
            detection_param.variance_encoded_in_target,
            detection_param.nms_param.eta,
            clip_box,
            objectness_score), det_out_dtype)

        self.params = self.params.flatten()
        self.params = self.params.view("<f4")

        self.max_detections = detection_param.keep_top_k + 1

    def getMaxDetections(self):
        return self.max_detections

class TF_DetectionOutput(DetectionOutput):
    _supported_IO_formats = [
        (LayerIOFormat.nhwc_compact(), LayerIOFormat.nhwc_compact(),
         LayerIOFormat.nchw_compact())
        ]

    def __init__(self, *args):
        """ The DetectionOutput layer does not exist in TensorFlow. The operation
            performed by DetectionOutput layer in TensorFlow is done with more
            primitive operations. In order to support the TF SSD architecures
            we introduce an DetectionOutput layer to replace the TF primitves.
            This DetectionOutput layer differs from the Caffe(plain DetectionOutput)
            in the number of input tensors it has only 2: mbox_loc and mbox_conf.
            The prior_box input tensor is a parameter because it is evaluated in
            TensorFlow and passd by data.
        """
        super().__init__(*args)

    def set_prior_boxes(self, prior_boxes):
        self.prior_boxes = PopulatedTensor(prior_boxes)
        self.prior_boxes.setLayout(Layouts.NHWC)

    def load_parameters(self, ssd_config):
        det_out_dtype = np.dtype("<i4, <i4, <i4, <f4, <i4, <i4, <i4, <f4, <i4, <f4, <i4")

        share_location = 1 if ssd_config.postprocessing_params.share_location else 0
        self.params= np.array(
            (ssd_config.postprocessing_params.num_classes,
             share_location,
             ssd_config.postprocessing_params.background_label_id,
             ssd_config.postprocessing_params.nms_params.iou_threshold,
             ssd_config.postprocessing_params.nms_params.max_detections_per_class,
             ssd_config.postprocessing_params.box_params.code_type,
             ssd_config.postprocessing_params.max_detections,
             ssd_config.postprocessing_params.nms_params.score_threshold,
             ssd_config.postprocessing_params.variance_encoded_in_target,
             ssd_config.postprocessing_params.nms_params.eta,
             ssd_config.postprocessing_params.box_params.clip_box),
            det_out_dtype)

        self.params = self.params.flatten()
        self.params = self.params.view("<f4")

        self.max_detections = ssd_config.postprocessing_params.max_detections + 1

