# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.

from collections import namedtuple
from enum import Enum

import Models.Layouts as Layouts

#The values in the EnclosureRequest have the following meaning:
#1. For the N, H, W axes the request is for the enclosure shape sizes to greater
#or equal to the original tensor siz and a mutliple of the EnclosureRequest value.
#2. For the C axis the request is for the enclosure shape size to be greater or equal
#to the EnclosureRequest value and greater or equal to the original tensor size.
EnclosureRequest = namedtuple("EnclosureRequest", "N C H W")
EnclosureRequest.__new__.__defaults__ = (1,) * len(EnclosureRequest._fields)

TensorShape = namedtuple('TensorShape', 'N C H W')

TENSOR_FORMAT = namedtuple("TensorFormat", "layout shape enclosure_shape")
class TensorFormat(TENSOR_FORMAT):
    def __new__(cls, layout, shape, enclosure_shape):
        self = super(TensorFormat, cls).__new__(cls, layout, TensorShape(*shape),
                                                   TensorShape(*enclosure_shape))
        return self

    @staticmethod
    def __roundup(x, mult):
        return (x + mult - 1) // mult * mult

    def is_compact(self):
        return self.shape == self.enclosure_shape

    @staticmethod
    def compute_enclosure_shape(tensor_shape, enclosure_request):
        return TensorShape(*[TensorFormat.__roundup(t_axis, enc_axis_req) for(t_axis, enc_axis_req)
                           in zip(tensor_shape, enclosure_request)])

    def get_enclosure_shape(self):
        return self.enclosure_shape

class LayerIOFormatPair(Enum):
    NCHW_to_NCHW_fullstride = 0
    NCHW_to_NHWC_fullstride = 1
    NCHW_to_NHCW_fullstride = 2

    NHWC_to_NCHW_fullstride = 3
    NHWC_to_NHWC_fullstride = 4
    NHWC_to_NHCW_fullstride = 5

    NHCW_to_NCHW_fullstride = 6
    NHCW_to_NHWC_fullstride = 7
    NHCW_to_NHCW_fullstride = 8

    NHWC_to_NHWC_W_stride   = 9
    NHCW_to_NHCW_C_stride   = 10

    NHCW_to_NHCW_C_stride_8 = 11
    NCHW_to_NCHW_H_stride_8 = 12

    NHCW_Cs8_to_NCHW_Hs8 = 13
    NCHW_Hs8_to_NHCW_Cs8 = 14

    NHWC_to_NHWC_compact = 15

IO_FORMAT = namedtuple("IO_FORMAT", "layout stridable_axes enclosure_request")
class LayerIOFormat(IO_FORMAT):
    def __new__(cls, layout, stridable_axes, enclosure_request=EnclosureRequest()):
        self = super(LayerIOFormat, cls).__new__(cls, layout, stridable_axes, enclosure_request)
        return self

    @classmethod
    def __fullstride(cls, layout):
        return cls(layout, (Layouts.Axes.N, Layouts.Axes.C, Layouts.Axes.H, Layouts.Axes.W))

    @classmethod
    def nchw_fullstride(cls):
        return cls.__fullstride(Layouts.NCHW)

    @classmethod
    def nhwc_fullstride(cls):
        return cls.__fullstride(Layouts.NHWC)

    @classmethod
    def nhcw_fullstride(cls):
        return cls.__fullstride(Layouts.NHCW)

    @classmethod
    def nhwc_w_stride(cls):
        return cls(Layouts.NHWC, (Layouts.Axes.W,))

    @classmethod
    def nhcw_c_stride(cls):
        return cls(Layouts.NHCW, (Layouts.Axes.C,))

    @classmethod
    def nhcw_c_stride_8(cls):
        nhcw_fs = cls.nhcw_fullstride()
        return cls(nhcw_fs.layout, nhcw_fs.stridable_axes, (EnclosureRequest(C=1, N=1, W=8, H=1)))

    @classmethod
    def nchw_h_stride_8(cls):
        nchw_fs = cls.nchw_fullstride()
        return cls(nchw_fs.layout, nchw_fs.stridable_axes, (EnclosureRequest(H=1, N=1, W=8, C=1)))

    @classmethod
    def nchw_compact(cls):
        return cls(Layouts.NCHW, ())

    @classmethod
    def nhwc_compact(cls):
        return cls(Layouts.NHWC, ())

    @classmethod
    def getLayerIOFormatPair(cls, format_pair):
        format_pairs = {
            LayerIOFormatPair.NCHW_to_NCHW_fullstride: (cls.nchw_fullstride(), cls.nchw_fullstride()),
            LayerIOFormatPair.NCHW_to_NHWC_fullstride: (cls.nchw_fullstride(), cls.nhwc_fullstride()),
            LayerIOFormatPair.NCHW_to_NHCW_fullstride: (cls.nchw_fullstride(), cls.nhcw_fullstride()),

            LayerIOFormatPair.NHWC_to_NCHW_fullstride: (cls.nhwc_fullstride(), cls.nchw_fullstride()),
            LayerIOFormatPair.NHWC_to_NHWC_fullstride: (cls.nhwc_fullstride(), cls.nhwc_fullstride()),
            LayerIOFormatPair.NHWC_to_NHCW_fullstride: (cls.nhwc_fullstride(), cls.nhcw_fullstride()),

            LayerIOFormatPair.NHCW_to_NCHW_fullstride: (cls.nhcw_fullstride(), cls.nchw_fullstride()),
            LayerIOFormatPair.NHCW_to_NHWC_fullstride: (cls.nhcw_fullstride(), cls.nhwc_fullstride()),
            LayerIOFormatPair.NHCW_to_NHCW_fullstride: (cls.nhcw_fullstride(), cls.nhcw_fullstride()),

            LayerIOFormatPair.NHWC_to_NHWC_W_stride : (cls.nhwc_w_stride(), cls.nhwc_w_stride()),
            LayerIOFormatPair.NHCW_to_NHCW_C_stride : (cls.nhcw_c_stride(), cls.nhcw_c_stride()),

            LayerIOFormatPair.NHCW_to_NHCW_C_stride_8 : (cls.nhcw_c_stride_8(), cls.nhcw_c_stride_8()),
            LayerIOFormatPair.NCHW_to_NCHW_H_stride_8 : (cls.nchw_h_stride_8(), cls.nchw_h_stride_8()),

            LayerIOFormatPair.NHCW_Cs8_to_NCHW_Hs8 : (cls.nhcw_c_stride_8(), cls.nchw_h_stride_8()),
            LayerIOFormatPair.NCHW_Hs8_to_NHCW_Cs8 : (cls.nchw_h_stride_8(), cls.nhcw_c_stride_8()),

            LayerIOFormatPair.NHWC_to_NHWC_compact: (cls.nhwc_compact(), cls.nhwc_compact()),
        }

        return format_pairs.get(format_pair)

    def can_produce(self, tensor_format):
        return self.can_consume(tensor_format)

    def can_consume(self, tensor_format):
        if self.layout != tensor_format.layout:
            return False

        for axis_i, axis in enumerate(self.layout):
            if tensor_format.shape[axis] != tensor_format.enclosure_shape[axis]:
                if axis_i == 0 or self.layout[axis_i - 1] not in self.stridable_axes:
                    return False

        for axis in self.layout:
            if axis != Layouts.Axes.C:
                if tensor_format.enclosure_shape[axis] % self.enclosure_request[axis] != 0:
                    return False
            else:
                if tensor_format.enclosure_shape[axis] < self.enclosure_request[axis]:
                    return False

        return True
