
# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.

from functools import reduce

import numpy as np
import Controllers.Adaptor as adaptor
from Controllers.Parsers.Parser.Hw import HwOp, HwConvolution, HwFC

from collections import OrderedDict

from Controllers.Parsers.Parser.Input import Input
from Controllers.Parsers.Parser.Output import Output
from Controllers.Optimizer import findClosestChild
from Controllers.Parsers.Parser.DetectionOutput import DetectionOutput, TF_DetectionOutput
from Controllers.Parsers.Parser.Bias import Bias
from Controllers.Parsers.Parser.Eltwise import Eltwise
from Controllers.Parsers.Parser.LRN import LRN
from Controllers.Parsers.Parser.Square import Square
from Controllers.Parsers.Parser.Pooling import Pooling
from Controllers.Parsers.Parser.Conversion import Conversion
from Controllers.Parsers.Parser.Layer import MangledName, OriginalName
from Controllers.Parsers.Parser.ReLU import ReLU
from Controllers.Parsers.Parser.crop import Crop
from Controllers.Parsers.Parser.Slice import Slice
from Controllers.Parsers.Parser.Softmax import Softmax
from Controllers.Parsers.Parser.Sigmoid import Sigmoid
from Controllers.Parsers.Parser.Convolution2D import Deconvolution, Convolution2D, ConvolutionDepthWise2D
from Controllers.Parsers.Parser.crop import Crop
from Controllers.Parsers.Parser.Concat import Concat
from Controllers.Parsers.Parser.Permute import PermuteFlatten
from Controllers.Parsers.Parser.ReshapeHWC import ReshapeHWC
from Controllers.Tensor import UnpopulatedTensor, Tensor
from Controllers import tf_ssd_config_pb2
import Models.Layouts as Layouts

from Controllers.EnumController import throw_error

from Models.EnumDeclarations import ErrorTable, PadStyle
from Controllers.GraphUtils import buildGraph, buildLayerLists
from Controllers.Optimizer import debug_graph
from Views.IRVisualize import drawGraph
import networkx as nx


"""
    This file is responsible for each framework-independent 'pass'.
    Thus, it should not rely on any prior framework knowledge, and
    not treat frameworks independently.
"""


def doWeightScaling(layer, out_data, per_channel=False, slicing = False):
    if slicing:
        idx, groups = slicing
        out_data = out_data[:,idx*out_data.shape[1]//groups:(idx+1)*out_data.shape[1]//groups,:,:]

    if per_channel:
        data_std_k = np.std(out_data, axis=(0, 2, 3))
        data_mean_k = np.mean(out_data, axis=(0, 2, 3))
    else:
        data_std_k = np.std(out_data)
        data_mean_k = np.mean(out_data)

    max_scale = 16384
    postOp = [x for x in layer.getPostOp() if isinstance(x, ReLU)]
    if (len(postOp) == 1) and hasattr(postOp[0], 'negativeSlope') and postOp[0].negativeSlope != 0:
        max_scale = min(max_scale, max_scale* abs(layer.post_op[0].negativeSlope))
    scale = np.clip(np.divide(1024 * np.ones(data_std_k.shape), 5 * (data_std_k + abs(data_mean_k))), 0.01, max_scale)
    print('Optimal scale for {0}: {1:.2f}'.format(layer.getName().stringifyOriginalName(),scale))

    w = layer.getWeights()
    w.data *= scale
    if layer.biasEnabled():
        b = layer.getBias()
        b.data *= scale
    layer.setScale(1 / scale)


def adaptHwWeights(parser, parsedLayers, input_range=(-1, 1), n_batch=1, per_channel=False):
    '''
    Idea: I want to adapt the output to the dinamics of the DPE accumulator
    the DPE accumulator takes fp16 input from the multiplicator and convert in into
    12q12 fixed point format. using 5 sigma confidence,  the 99.9999% of the output
    data is inside the range [-5 sigma, + 5 sigma]. This will adapt nicely the dynamics

    Standard Convolution/FC operation (* may be a convolution or a matrix multiplication)
    Out = in * W + b
    Out' = in * W' + b' = Out * scale = in * (W*scale) + b*scale

    Variance operator is linear to data scaling
    var{Out'}  = scale * var{Out}

    optimal_range = [-2048.xx; 2047.xx] (12 bit of integer part)
    max{optimal_range} = 5*var{Out'} = 5*scale*var{Out}

    scale = max{optimal_range} / 5 / var{Out}

    This operation can be done (in conv) per output channel
    '''
    print('Evaluating input and weigths for each hw layer')
    if parser.getType() == 'Caffe':
        for layer in parsedLayers:
            # TODO: do it for FC!
            # TODO: implement per channel scaling
            if (isinstance(layer, HwConvolution) or isinstance(layer, HwFC)) and layer.hasWeights():

                outBlobName = layer.outputTensors[0].name.stringifyOriginalName()
                out_data = parser.get_layer_data(outBlobName)
                doWeightScaling(layer, out_data, per_channel, layer.sliced)

    if parser.getType() == 'TensorFlow':
        hw_layers = [layer for layer in parsedLayers if (isinstance(layer, HwConvolution) or isinstance(layer, HwFC)) and layer.hasWeights()]
        outBlobNames = [layer.outputTensors[0].name.stringifyOriginalName() for layer in hw_layers]
        out_data = parser.get_layers_data(outBlobNames)
        for idx, layer in enumerate(hw_layers):
            doWeightScaling(layer, out_data[idx], per_channel, layer.sliced)


    print('--------------------------------------')

    return parsedLayers

def selectInputOutputLayout(parsedLayers, input_layout, output_layout):
    for layer in parsedLayers:
        if input_layout is not None and isinstance(layer, Input):
            layer.force_layout(input_layout)
        if output_layout is not None and isinstance(layer, Output):
            layer.force_layout(output_layout)

    return parsedLayers

def flattenFCLayers(parsedLayers):
    """
        Optimization pass to convert 2D InnerProducts to 1D equivilants.
    """

    from Controllers.Parsers.Parser.InnerProduct import InnerProduct

    for layer in parsedLayers:
        if isinstance(layer, InnerProduct):
            s = layer.getInputTensors()[0].getShape()
            if len(s) - 1 == s.count(1):    # If there is only 1 Dimension
                print("1D")
            else:
                print("2D")
                """
                TO BE IMPLEMENTED
                """
                print("WARNING: Unimplemented 2D FC")

    return parsedLayers

def reinsertOutputOps(parsedLayers):
    # TODO: This is a copy of the one in Caffe. Come up with a proper answer to whose responsibility it is

    # Find all tensors that are not consumed by anybody
    tensorNames = OrderedDict()
    for layer in parsedLayers:
        for tensor in layer.getInputTensors():
            if tensor.getName().stringifyName() not in tensorNames:
                tensorNames[tensor.getName().stringifyName()] = ([0, 0], tensor, [])

            tensorNames[tensor.getName().stringifyName()][0][0] += 1

        for tensor in layer.getOutputTensors():
            if tensor.getName().stringifyName() not in tensorNames:
                tensorNames[tensor.getName().stringifyName()] = ([0, 0], tensor, [])

            tensorNames[tensor.getName().stringifyName()][0][1] += 1
            tensorNames[tensor.getName().stringifyName()][2].append(layer)

    for tensorName, tensorValue in tensorNames.items():
        consumersAndProducers, tensor, producers = tensorValue
        consumers = consumersAndProducers[0]
        if consumers == 0:
            x = Output('Output', [tensor.getName()], [])
            x.setInputTensors([tensor])
            x.setOutputTensors([])
            x.loadInputTensorSizes([tensor.getShape()])
            x.loadOutputTensorSizes([])

            assert(len(producers) == 1)
            if isinstance(producers[0], DetectionOutput):
                x.enableDetectionOutput()

            parsedLayers.append(x)

    return parsedLayers


def cropNetwork(parsedLayers, arguments):
    """
        Apply output-node-name and input-node-name
    """

    if arguments.output_node_name is None and arguments.input_node_name is None:
        return parsedLayers

    exit_early = False
    start_late = True if arguments.input_node_name is not None else False

    from Controllers.GraphUtils import buildGraph
    from Controllers.Optimizer import debug_graph
    import networkx as nx

    g = buildGraph(parsedLayers)
    # print("========== BEFORE ===========")
    # debug_graph(g)
    croppedParsedLayers = []

    for index, child in enumerate(nx.lexicographical_topological_sort(g)):
        layer = g.node[child]['ref']

        if index == 0:
            first_layer = layer

        if layer.name.stringifyOriginalName() == arguments.input_node_name:
            start_late = False
            print("Start Layer:", layer)

            pred = list(g.predecessors(child))[0]
            p_node = g.node[pred]['ref']
            p_node = Input("Input", first_layer.getInputTensorNames(), p_node.getOutputTensorNames())

            p_node.setInputTensors(first_layer.getInputTensors())  # UnpopulatedTensor(first_layer.getInputTensorNames()[0])
            p_node.loadInputTensorSizes({})

            p_node.setOutputTensors([layer.getInputTensors()[0]])
            p_node.loadOutputTensorSizes({layer.getInputTensors()})

            print("Source: ", p_node)
            croppedParsedLayers.append(p_node)

        if not exit_early and not start_late:
            croppedParsedLayers.append(layer)
        else:
            if exit_early:
                pred = list(g.predecessors(child))[0]
                pred_node = g.node[pred]['ref']
                print(pred_node.name.stringifyOriginalName(), "==", arguments.output_node_name)

                if exit_early is True and pred_node.name.stringifyOriginalName() == arguments.output_node_name and layer.getInPlace():
                    # Only allow in-place operations to be attached from this point on
                    croppedParsedLayers.append(layer)

        if layer.name.stringifyOriginalName() == arguments.output_node_name:
            # Break the network early as instructed by user
            exit_early = True

    if arguments.output_node_name is not None:
        reinsertOutputOps(croppedParsedLayers)


    g = buildGraph(parsedLayers)
    # print("========== AFTER ===========")
    # debug_graph(g)
    # print("========== ===== ===========")

    return croppedParsedLayers

def forceTensorLayout(parsedLayers):
    for layer in parsedLayers:
        if layer.isHW:
            i_l, o_l = layer.getPingPong()[2]
            from Models.Layouts import NCHW, NHCW

            if i_l == 'P':
                layer.getInputTensors()[0].setLayout(NCHW)
            else:
                layer.getInputTensors()[0].setLayout(NHCW)

            if o_l == 'P':
                layer.getOutputTensors()[0].setLayout(NCHW)
            else:
                layer.getOutputTensors()[0].setLayout(NHCW)


def Slice_with_port_crop_layers(parsedLayers):

    g = buildGraph(parsedLayers)

    for child in nx.lexicographical_topological_sort(g):
        layer = g.node[child]['ref']
        if isinstance(layer, Slice):
            slice_links = []
            offset_counter = 0
            succ = list(g.successors(child))
            pred = list((g.predecessors(child)))

            assert len(pred) == 1, "Slice not supported to have >1 predecessors"

            for ot in layer.getOutputTensors():
                isContinuing = False
                for sc in succ:
                    after = g.node[sc]['ref']
                    for it in after.getInputTensors():
                        if(ot.getName().stringifyOriginalName() == it.getName().stringifyOriginalName()):
                            slice_links.append([after, layer, ot, offset_counter, isContinuing])
                            isContinuing = True

                offset_counter += ot.getShape()[1]

            pre_layer = g.node[pred[0]]['ref']
            name_id = 0

            # Insert Crop Layers
            for sl in slice_links:
                s_node = sl[0]
                node = sl[1]
                node_out_tensor = sl[2]
                offset = sl[3]
                already_created = sl[4]
                input_tensor = pre_layer.getOutputTensors()[0]

                crop_in = pre_layer.getOutputTensors()[0].getName().stringifyName()
                crop_in_name = "crop_group_{0}_".format(name_id)

                next_name = node.getName().stringifyName()
                prev_name = pre_layer.getName().stringifyName()
                after_next_name = s_node.getName().stringifyName()

                if (already_created == False):
                    #Create the Crop Layer
                    crop = Crop(crop_in_name, [pre_layer.getOutputTensors()[0].getName()],
                                [node_out_tensor.getName()])
                    crop.setInputTensorsAllFields([pre_layer.getOutputTensors()[0]])
                    crop.setOutputTensorsAllFields([node_out_tensor])
                    crop.setOffset(np.array([0, 0, offset]))
                    #Port the Crop in the Graph
                    g.add_node(crop_in_name, type="OP", ref=crop)
                    g.add_edge(prev_name, crop_in_name)
                    g.add_edge(crop_in_name, after_next_name)
                    g.remove_edge(next_name, after_next_name)
                    name_id = name_id + 1

            g.remove_edge(prev_name, next_name)
            g.remove_node(next_name)

    parsedLayers = buildLayerLists(g)

    return parsedLayers


def compute_tiled_shape(tensors, concat_axis):
    tiled_shape = [0, 0, 0, 0]

    for tensor_idx, tensor in enumerate(tensors):
        t_shape = tensor.getShape()
        t_enc_shape = tensor.getTopEncloserRecursive().getShape()

        for axis_idx, axis in enumerate(tiled_shape):
            if axis_idx == concat_axis:
                # If a tensor that enters a Concat has an enclosure with the concat
                # axis size greater than the size of the original tensor it means
                # that this tensor is used by another layer besides the Concat layer
                # and that layer requires an enclosure for the concat axis.
                # This case is not supported at the moment.
                assert t_shape[axis_idx] == t_enc_shape[axis_idx]

                tiled_shape[axis_idx] = tiled_shape[axis_idx] + t_shape[axis_idx]
            else:
                tiled_shape[axis_idx] = max(tiled_shape[axis_idx], t_enc_shape[axis_idx])

    return tiled_shape


def implicitConcatBinding(parsedLayers):
    """
        In any concat layers, contain the inputs inside of the outputs.
        This will allow the myriad to write the preceding layers in the
        correct location, implicitly performing a concat.
    """

    from Controllers.Parsers.Parser.Concat import Concat

    for layer in parsedLayers:
        if isinstance(layer, Concat):
            input_tensors = layer.getInputTensors()
            output_tensor = layer.getOutputTensors()[0]
            output_tensor_original_enc = output_tensor.getTopEncloserRecursive().getShape()

            bpp = 2

            inputs_tiled_shape = compute_tiled_shape(input_tensors, layer.getAxis())

            final_concat_enc_shape = tuple(
                [max(out_enc_axis, tiled_enc_axis) for out_enc_axis, tiled_enc_axis in
                 zip(output_tensor_original_enc, inputs_tiled_shape)])

            if final_concat_enc_shape != output_tensor_original_enc:
                final_enc = UnpopulatedTensor(final_concat_enc_shape)
                final_enc.setName(MangledName(OriginalName(
                    layer.getName().stringifyName() + "_final_enclousre")))
                final_enc.setLayout(output_tensor.getLayout())
                output_tensor.place(final_enc, (0,0,0,0))

            concat_output_size = np.prod(output_tensor.getTopEncloserRecursive().getShape()) * bpp

            concat_axis_offset = 0
            for in_tensor in input_tensors:
                corner = tuple([0 if axis != layer.getAxis() else concat_axis_offset
                          for axis in range(0, len(in_tensor.layout))])

                in_tensor.getTopEncloserRecursive().place(
                    output_tensor.getTopEncloserRecursive(), corner)

                for c_layer in parsedLayers:
                    if isinstance(c_layer, HwOp) and in_tensor in c_layer.getOutputTensors():
                        if in_tensor.getLayout() == Layouts.NHCW:
                            hw_concat_offset = output_tensor.getTopEncloserRecursive().getShape()[Layouts.Axes.W] * \
                                               concat_axis_offset * bpp
                        elif in_tensor.getLayout() == Layouts.NCHW:
                            hw_concat_offset = output_tensor.getTopEncloserRecursive().getShape()[Layouts.Axes.W] * \
                                               output_tensor.getTopEncloserRecursive().getShape()[Layouts.Axes.H] * \
                                               concat_axis_offset * bpp
                        else:
                            assert False, "Bad NCE Layer layout."

                        c_layer.setConcatParameters(hw_concat_offset,
                                                    concat_output_size,
                                                    output_tensor.getTopEncloserRecursive().getShape()[Layouts.Axes.C])

                concat_axis_offset += in_tensor.getShape()[layer.getAxis()]

            layer.setImplicit()

    return parsedLayers


def serializeNewFmt(parsedLayers, arguments, myriad_conf, input_data):
    """
        Serializes the output/IR of the new parser into a representation
        that is readable by myriad (blob/graphFile)
    """

    # Find the input and output tensors
    inputTensors = set()
    outputTensors = set()
    outputLayers = set()

    for layer in parsedLayers:
        if type(layer) == Input:
            for tensor in layer.getOutputTensors():
                inputTensors.add(tensor)

        if type(layer) == Output:
            for tensor in layer.getInputTensors():
                outputTensors.add((tensor, layer.isDetectionOutput()))
                outputLayers.add(layer)

    # Find which outputTensor is closest to output_node_name

    from Controllers.Optimizer import buildGraph, debug_graph
    g = buildGraph(parsedLayers)

    # debug_graph(g)

    # If output node provided, select one particular output node.
    # e.g. split
    if arguments.output_node_name:
        selected_output = None
        for layer in parsedLayers:
            if layer.name.stringifyOriginalName() == arguments.output_node_name:
                selected_output = layer
        if selected_output is not None:
            closestLayer = findClosestChild(g, selected_output, outputLayers)

            outputTensors = set()
            for tensor in closestLayer.getInputTensors():
                outputTensors.add((tensor, layer.isDetectionOutput()))
        else:
            # output_node_name not found in graph.
            # in this case, leave the outputTensors unmodified and let
            # the NetworkEmulator object select the first tensor from this list
            pass

    # Convert to list for simpler manipulation
    inputTensors = list(inputTensors)
    outputTensors = list(outputTensors)

    assert(len(outputTensors) >= 1)

    print('# Network Input tensors', [tensor.getName().stringifyName() for tensor in inputTensors])
    print('# Network Output tensors', [tensor.getName().stringifyName() for tensor, _ in outputTensors])

    adp = adaptor.Adaptor([input_data], inputTensors, outputTensors)
    adp.transform(parsedLayers)     # TODO: DFS Transform?
    net, blob = adp.serialize(arguments, myriad_conf)

    return net, blob, parsedLayers


def compatibilityPadding(parsedLayers):

    from Controllers.Parsers.Parser.Convolution2D import Convolution2D
    from Controllers.Tensor import UnpopulatedTensor

    for layer in parsedLayers:
        if isinstance(layer, Convolution2D) and not isinstance(layer, HwOp):
            if True:  # TODO: Enable only on specific implementation
                """
                Gets an offset for a buffer in the work buffer. Relative to the start of the blob file.
                                +----------+
                                |          |
                                |   pad    |
                outputPointer+---> +----------+
                                |          |
                                |  buffer  |
                                |          |
                                +----------+
                                |          |
                                |   pad    |
                                +----------+
                Note: The buffer is overallocated for each stage, because convolutions write
                outside the buffers.
                The maximum pad is computed for RADIX_MAX, stride 1, and SAME padding. (i.e
                the kernel is centered from the first point in the image)
                Let's take radix = 3. For the first output we add up the first two elements
                of the first two rows from the input, of course, multiplied by the kernel weights.
                So for the last kernel (bottom-right), the input has an offset of 1 row and 1 column,
                which is radix/2 * row_width + radix/2. (In our algorithm this means it will write
                starting from -(radix/2 * row_width + radix/2))
                If we put the first kernel (upper-left) to overalp the first point in the image,
                we will get the result for the radix/2 * row_width + radix/2 output element.
                (In our algorithm it will write starting from (radix/2 * row_width + radix/2),
                so it will write outside with the same amount of elements)
                Therefore, the maximum offset has to be radix/2 * row_width + radix/2.
                """

                ot = layer.getOutputTensors()[0]
                ot_shape = ot.getShape()

                pad = layer.getKernelSize()[0]

                pad = (pad // 2) + 1

                padded_shape = tuple((
                    ot_shape[0] + 0,
                    ot_shape[1] + 0,
                    ot_shape[2] + 2 * pad,
                    ot_shape[3] + 0,
                ))

                ot.setDatatype(np.float16)  # TODO: This should be dynamically set
                padding = UnpopulatedTensor(padded_shape)
                padding.setName(ot.getName() + "_pad")
                padding.setLayout(ot.getLayout())
                padding.setDatatype(ot.getDatatype())

                ot.place(padding, (0, 0, pad, 0))

    return parsedLayers

def squashInPlaceLayers(parsedLayers):

    g = buildGraph(parsedLayers)
    # print("Inplace layers squashing")

    contraction_list = []

    # N - inPlaceNode
    # succ - children of that node
    # pred - parent node
    for n in g.node:
        if g.node[n]['ref'].getInPlace():
            pred = list(g.predecessors(n))
            if len(pred) == 1:
                contraction_list.append((n, pred[0]))
    # Fuse and Remove
    for n, pred in contraction_list:
        # print("Place {} {} in {} {}. Offset: {}". format(
        #                 g.node[pred]['ref'].getOutputTensors()[0].getTopEncloserRecursive().name.stringifyName(), g.node[pred]['ref'].getOutputTensors()[0].getTopEncloserRecursive().shape,
        #                 g.node[n]['ref'].getOutputTensors()[0].name.stringifyName(), g.node[n]['ref'].getOutputTensors()[0].shape, (0, 0, 0, 0)))
        g.node[pred]['ref'].getOutputTensors()[0].getTopEncloserRecursive().place(g.node[n]['ref'].getOutputTensors()[0], (0, 0, 0, 0))
        g.node[n]['ref'].getOutputTensors()[0].setDatatype(np.float16)

    return buildLayerLists(g)

def placeCroppedTensor(parsedLayers):
    'Implicit placing of cropped tensor into the parent one'
    from Controllers.Parsers.Parser.crop import Crop
    layers = parsedLayers
    for layer in layers:
        if isinstance(layer, Crop) and layer.getImplicit():
            assert(len(layer.getInputTensors())==1)
            assert(len(layer.getOutputTensors())==1)

            offset = (0,layer.crop_offset[0], layer.crop_offset[1], layer.crop_offset[2])

            # print("Place {} {} in {} {}. Offset: {}". format(
            #             layer.getOutputTensors()[0].getTopEncloserRecursive().name.stringifyName(),layer.getOutputTensors()[0].getTopEncloserRecursive().shape,
            #             layer.getInputTensors()[0].getTopEncloserRecursive().name.stringifyName(), layer.getInputTensors()[0].getTopEncloserRecursive().shape, offset))

            layer.getOutputTensors()[0].getTopEncloserRecursive().place(layer.getInputTensors()[0].getTopEncloserRecursive(), offset)

    return parsedLayers

def breakUpInnerLRN(parsedLayers):
    """
    Convert from:
    <  X  >
       |
    < LRN >
       |
    <  Y  >

    To
        <   X   >
         /     \
        /       \
    < Square >   \
        |        |
    < AvPool >   |
        |        |
    < InnerLRN > |
          \      |
        < Elt Prod >

    Only Valid for Inner LRN
    """

    g = buildGraph(parsedLayers)
    from Controllers.Optimizer import debug_graph

    # for n in g.node:
    for n in nx.lexicographical_topological_sort(g):
        layer = g.node[n]['ref']
        if isinstance(layer, LRN) and layer.getType() == LRN.Type.WITHIN:

            n = layer.name.stringifyOriginalName()
            itOriginal = layer.getInputTensors()
            otSample = layer.getOutputTensors()[0]

            # Square

            sq_out = UnpopulatedTensor(otSample.shape)
            sq_out.setName(MangledName(OriginalName(n + "_sq_out")))

            sq = Square(n + "_sq", layer.getInputTensorNames(), [sq_out.getName()])
            sq.setInputTensors(layer.getInputTensors())
            sq.setInputTensorNames(layer.getInputTensorNames())
            sq.loadInputTensorSizes(layer.getInputTensorSizes())

            sq.setOutputTensors([sq_out])
            sq.setOutputTensorNames([sq_out.name])
            sq.loadOutputTensorSizes([sq_out.shape])

            g.add_node(sq.name.stringifyName(), type="OP", ref=sq)
            for p in list(g.predecessors(layer.name.stringifyName())):
                g.add_edge(p, sq.name.stringifyName())
                g.remove_edge(p, layer.name.stringifyName())

            # Average Pool

            av_out = UnpopulatedTensor(otSample.shape)
            av_out.setName(MangledName(OriginalName(n + "_av_out")))

            av = Pooling(n + "_av", [sq_out.getName()], [av_out.getName()])
            av.loadType(Pooling.Type.AVE)

            av.loadKernelSize(layer.getSquareKernelSize(), layer.getSquareKernelSize())
            av.loadStrideSize(1, 1)
            pad_size = (layer.getSquareKernelSize() - 1) // 2
            av.loadPadding(pad_size, pad_size, pad_size, pad_size)

            av.setInputTensors([sq_out])
            av.setInputTensorNames([sq_out.name])
            av.loadInputTensorSizes([sq_out.shape])

            av.setOutputTensors([av_out])
            av.setOutputTensorNames([av_out.name])
            av.loadOutputTensorSizes([av_out.shape])

            g.add_node(av.name.stringifyName(), type="OP", ref=av)
            g.add_edge(sq.name.stringifyName(), av.name.stringifyName())

            # LRN Partial
            # (1 + alpha * prev) ^ -beta
            layer.setInputTensors([av_out])
            layer.setInputTensorNames([av_out.name])
            layer.loadInputTensorSizes([av_out.shape])
            layer.loadType(LRN.Type.WITHIN_PARTIAL)  # Do not re-enter this check
            layer.loadSquareKernelSize(1)
            g.add_edge(av.name.stringifyName(), layer.name.stringifyName())

            # Elt Prod
            ep_out = UnpopulatedTensor(otSample.shape)
            ep_out.setName(MangledName(OriginalName(n + "_ep_out")))

            ep = Eltwise(n + "_ep", layer.getOutputTensorNames(), [ep_out.getName()])

            ep.loadType(Eltwise.Type.WPROD)

            ep_in = list(layer.getOutputTensors())
            assert len(itOriginal) == 1, "Myriad does not support Within LRN where there is multiple input to the layer"
            ep_in.extend(itOriginal)

            ep.setInputTensors(ep_in)
            ep.setInputTensorNames([l.getName() for l in ep_in])
            ep.loadInputTensorSizes([l.getShape() for l in ep_in])

            ep.setOutputTensors([ep_out])
            ep.setOutputTensorNames([ep_out.name])
            ep.loadOutputTensorSizes([ep_out.shape])

            g.add_node(ep.name.stringifyName(), type="OP", ref=ep)

            for s in list(g.successors(layer.name.stringifyName())):
                g.node[s]['ref'].setInputTensors([ep_out])
                g.node[s]['ref'].setInputTensorNames([ep_out.name])
                g.node[s]['ref'].loadInputTensorSizes([ep_out.shape])
                g.remove_edge(layer.name.stringifyName(), s)
                g.add_edge(ep.name.stringifyName(), s)

            g.add_edge(layer.name.stringifyName(), ep.name.stringifyName())

    parsedLayers = buildLayerLists(g)

    return parsedLayers


def overrideSoftwareSetting(parsedLayers, schedule):
    for l in parsedLayers:
        lname = l.name.stringifyOriginalName()
        if lname in schedule.ordered_dict:
            layout = schedule.ordered_dict[lname][1]
            if layout is False or layout not in ["CHW", "HCW", "HWC"]:
                continue
            else:
                l.force_layout(layout)

    return parsedLayers


def forceSymmetricalPadding(parsedLayers):
    for l in parsedLayers:
        if type(l) in [Convolution2D, ConvolutionDepthWise2D, Deconvolution]:
            l.force_symmetrical_padding()

    return parsedLayers



def transformForHwImplementation(parsedLayers):
    get_transform_list = lambda layer: [layer] if type(layer) not in [Convolution2D] \
                                        else layer.implementAsPadPlusConv()
    return [new_layer for layer in parsedLayers for new_layer in get_transform_list(layer)]


def splitGroupDeconvolution(parsed_layers):
    get_group_list = lambda layer: [layer] if not isinstance(layer, Deconvolution) \
                                        else layer.getSplitGroupDeconvolution()
    return [new_layer for layer in parsed_layers for new_layer in get_group_list(layer)]

def splitGroupConvolutionForShaves(parsed_layers):
    get_group_list = lambda layer: [layer] if not isinstance(layer, Convolution2D) \
                                        else layer.getSplitGroupConvolutionForShaves()
    return [new_layer for layer in parsed_layers for new_layer in get_group_list(layer)]

def caffeCropAdjustment(parsed_layers):
    """
        Removes the layers and tensors that are the second input to a Crop layer if
        the layer/tensor is only connected to the Crop layer.

        In Caffe the Crop layer has 2 inputs - 1st the tensor to be cropped and the
        2nd one give the dimensions to which to crop the 1st input.
        In MvTensor the Crop layer has a single input and the dimensions to which to
        crop the input are given by the dimensions of the output tensor. In the compiler
        the dimensions of the output tensor are taken directly from the reference
        framework (Caffe) hence the second input to the Crop layer is not necessary
        in our framework.
    """

    graph = buildGraph(parsed_layers)

    for current_node_name in nx.lexicographical_topological_sort(graph):
        current_node = graph.node[current_node_name]
        if current_node['type'] == 'OP' and isinstance(current_node['ref'], Crop):
            crop_parents = list(graph.predecessors(current_node_name))
            assert len(crop_parents) == 2, "Caffe waits for 2 bottom tensors in Crop."

            #Sometimes graph.predecessors(crop)[1] returns bottom[0] of Crop
            for previous_node in crop_parents:
                if graph.node[previous_node]['ref'].getOutputTensorNames()[0] == current_node['ref'].getInputTensorNames()[1]:
                    possible_layer_to_remove = graph.node[previous_node]['ref'].name.stringifyName()

            layers_to_remove = []
            while (len(list(graph.successors(possible_layer_to_remove)))) == 1:
                layers_to_remove.append(possible_layer_to_remove)
                if len(list(graph.predecessors(possible_layer_to_remove))) == 1:
                    possible_layer_to_remove = list(graph.predecessors(possible_layer_to_remove))[0]
                else:
                    throw_error(ErrorTable.UnsupportedOperation, "Architecture of Crop's second bottom tensor")

            for i in layers_to_remove:
                graph.remove_node(i)

            crop_layer = current_node['ref']
            crop_layer.inputTensorNames = [crop_layer.inputTensorNames[0]]
            crop_layer.setInputTensors([crop_layer.getInputTensors()[0]])
            crop_layer.loadInputTensorSizes([tensor.getShape() for tensor in crop_layer.getInputTensors()])

    return buildLayerLists(graph)

def solveRefineDetSpeicificDetectionOutput(parsedLayers):
    from Controllers.Parsers.Parser.DetectionOutput import DetectionOutput

    for layer in parsedLayers :
        if(len(layer.inputTensors) > 4 and type(layer) is DetectionOutput):
            #This is how  we will assume we are in the RefineDet implemetation of the DetectionOutput layer
            #the 4th input is assumed to be the anchor refinement modules confidence vector. Since the current runtime
            #limits us to 1 input tensor (the 3 input tensor version is ha hack too, since we say they are "weights"and "biases")
            #So, we are constrained to "Concat" some of the inputs.  We will chose to concat the 2 confidences and the 2 location vectors.
            #The PrioBox will remain the same. Since the runtime framework will only be properly aware of 3 """Inputs"""
            #We will pass the location of the 2 ""extra"" inputs in the parameters buffer.

            #TLDR version: big hack to fit 5 inputs into 3. (No guarantee it will work if not heavily maintained.)
            #              also, if the behaviour is changed here, make sure to double-check with the

            #naming convention: odm - Object Detection Module.
            #                   arm - Anchor Refinement Module
            # odm is basically the standard SSD input. Will use this name to diferentiate in RefineDet context

            odm_loc = layer.getInputTensors()[0]
            odm_conf = layer.getInputTensors()[1]

            arm_conf = layer.getInputTensors()[3]
            arm_loc = layer.getInputTensors()[4]

            #will make the bold assumption that these are flattened layers, which only have data in the channels dimention
            #this makes the "concat" simple, since they only need to be placed sequentially in memory
            odm_shape = odm_loc.shape
            arm_shape = arm_loc.shape
            assert( (odm_shape[2] == 1) and (odm_shape[3] == 1) and (arm_shape[2] == 1) and (arm_shape[3] == 1) )
            odm_shape = odm_conf.shape
            arm_shape = arm_conf.shape
            assert( (odm_shape[2] == 1) and (odm_shape[3] == 1) and (arm_shape[2] == 1) and (arm_shape[3] == 1) )

            #also, they should be compact tensors
            assert( odm_loc.shape == odm_loc.getTopEncloserRecursive().shape )
            assert( odm_conf.shape == odm_conf.getTopEncloserRecursive().shape )
            assert( arm_loc.shape == arm_loc.getTopEncloserRecursive().shape )
            assert( arm_conf.shape == arm_conf.getTopEncloserRecursive().shape )

            #also same layouts
            assert( odm_loc.getLayout() == arm_loc.getLayout() )
            assert( odm_conf.getLayout() == arm_conf.getLayout() )

            odm_loc_size = odm_loc.shape[1]
            arm_loc_size = arm_loc.shape[1]
            odm_conf_size = odm_conf.shape[1]
            arm_conf_size = arm_conf.shape[1]

            loc_data_size = odm_loc.shape[1] + arm_loc.shape[1]
            conf_data_size = odm_conf.shape[1] + arm_conf.shape[1]

            loc_tensor = UnpopulatedTensor( (1,loc_data_size,1,1) )
            loc_tensor.setLayout( odm_loc.getLayout() ) #same as arm_loc, since we asserted it
            loc_tensor.setName(MangledName(OriginalName("detectionOutput_odm_arm_loc")))

            conf_tensor = UnpopulatedTensor( (1,conf_data_size,1,1) )
            conf_tensor.setLayout( odm_conf.getLayout() ) #same as arm_conf, since we asserted it
            conf_tensor.setName(MangledName(OriginalName("detectionOutput_odm_arm_conf")))

            #place the tensors. For the sake of not diverging too much from standard SSD behaviour, will keep ODM's at the start
            odm_loc_offset = (0,0,0,0)
            arm_loc_offset = (0,odm_loc_size,0,0)

            odm_conf_offset = (0,0,0,0)
            arm_conf_offset = (0,odm_conf_size,0,0)

            odm_loc.place(loc_tensor, odm_loc_offset)
            arm_loc.place(loc_tensor, arm_loc_offset)

            odm_conf.place(conf_tensor,odm_conf_offset)
            arm_conf.place(conf_tensor,arm_conf_offset)

    return parsedLayers

def _tf_ssd_gen_mbox_loc_patch(parsed_layers, mbox_loc_names):
    """ Generate a dict of patch layers for SDD box location detection heads.

        Note: The shapes in this description are given as [N, H, W, C].

        The location(mbox_loc) detection heads in TensorFlow consist of a convolution
        followed by a reshape. The output tensor of the reshape has the shape:
        [batch_size, num_boxes, 1, 4] and the layout is NHWC.

        The MvTensor DetectionOutput layer expects a box location tensor with the
        same layout as TensorFlow but flattened towards the C axis i.e: box location
        with shape [batch_size, 1, 1, num_boxes * 4] and the layout is NHWC.
        In order to achieve this the Reshape layers will be replaced with PermuteFlatten
        layers with the permutation set to Identity and a Concat layer will be inserted.

        Inputs:
            parsed_layers - all the layers in the network.
            mbox_loc_names - names of the Tensors containing the box locations
                         for the SSD detection heads that enter the patch.

        Returns: (parsed_layers - updated list of parsed layers
                  mbox_loc - output tensor)
    """

    # Search for the SSD detection head Tensor in the output tensor list of the
    # layers.
    mbox_detection_heads = []
    for mbox_name in mbox_loc_names:
        det_head_found = False
        for layer in parsed_layers:
            output_tensor_names = [name.stringifyOriginalName()
                                   for name in layer.getOutputTensorNames()]
            if mbox_name in output_tensor_names:
                mbox_detection_heads.append(layer)
                det_head_found = True
                # There should not be another layer that outputs the same tensor
                break

        assert det_head_found, "Could not find mbox tensor: {0}".format(mbox_name)

    # Create PermuteFlatten replacement layers.
    concat_input_tensors = []
    perm_flatten_layers = []
    for det_head in mbox_detection_heads:
        perm_flatten_shape = (1,
                              reduce(lambda x, y: x * y, det_head.getOutputTensors()[0].shape),
                              1,
                              1)
        mbox_perm_flatten_output = UnpopulatedTensor(perm_flatten_shape)
        mbox_perm_flatten_output.setName(MangledName(
            OriginalName(det_head.getOutputTensorNames()[0].stringifyOriginalName())))
        concat_input_tensors.append(mbox_perm_flatten_output)
        mbox_perm_flatten = PermuteFlatten("PERM_FLATTEN" + det_head.getName().stringifyOriginalName(), [], [])
        mbox_perm_flatten.setInputTensorsAllFields(det_head.getInputTensors())
        mbox_perm_flatten.setOutputTensorsAllFields([mbox_perm_flatten_output])
        mbox_perm_flatten.reference_layout = Layouts.NHWC
        mbox_perm_flatten.permutation = (0, 1, 2, 3)
        perm_flatten_layers.append(mbox_perm_flatten)


    C_axis_concat_size = reduce(lambda x, y: x + y,
                                [t.shape[Layouts.Axes.C] for t in concat_input_tensors])
    mbox_concat_shape = (1, C_axis_concat_size, 1, 1)
    mbox_concat_output = UnpopulatedTensor(mbox_concat_shape)
    mbox_concat_output.setName(MangledName(OriginalName("mbox_loc")))
    mbox_concat = Concat("mbox_loc", [], [])
    mbox_concat.setInputTensorsAllFields(concat_input_tensors)
    mbox_concat.setOutputTensorsAllFields([mbox_concat_output])
    # This concat axis, normally, should be H because the Layout should be NHWC but
    # the implict concat  mechanism might not like it and because the concat inputs
    # should be flattend on the C axis it won't matter which axis is slected hence
    # use C axis.
    mbox_concat.loadConcatAxis(Layouts.Axes.C)

    # Remove mbox_loc reshape layers.
    parsed_layers = [layer for layer in parsed_layers if layer not in mbox_detection_heads]

    # Add PermuteFlatten layers
    parsed_layers = parsed_layers + perm_flatten_layers

    parsed_layers.append(mbox_concat)

    return (parsed_layers, mbox_concat_output)

def _tf_ssd_gen_mbox_conf_patch(parsed_layers, mbox_conf_names, det_out_config):
    """ Generate a dict of patch layers for SDD box confidence/score detection heads.

        Note: The shapes in this description are given as [N, H, W, C].

        The confidence(mbox_conf) detection heads in TensorFlow consist of a convolution
        followed by a reshape. The output tensor of the reshape has the shape:
        [batch_size, num_boxes, 1, num_classes] and the layout is NHWC.

        The MvTensor DetectionOutput layer expects a box confidence tensors with the
        same layout as TensorFlow but flattened towards the C axis i.e: box confidence
        with shape [batch_size, 1, 1, num_boxes * 4] and the layout is NHWC.
        In order to achieve this the Reshape layers will be replaced with PermuteFlatten
        layers with the permutation set to Identity. The flattend outputs will be
        concatenated using a Concat layer. The output of the Concat layer will be
        reshaped in place via a ReshapeHWC layer and then passed through a Sofmax layer.

        Inputs:
            parsed_layers - all the layers in the network.
            mbox_conf_names - names of the Tensors containing the box locations
                         for the SSD detection heads that enter the patch.

        Returns: (parsed_layers - updated list of parsed layers
                  mbox_conf - output tensor)
    """

    # Search for the SSD detection head Tensor in the output tensor list of the
    # layers.
    mbox_detection_heads = []
    for mbox_name in mbox_conf_names:
        det_head_found = False
        for layer in parsed_layers:
            output_tensor_names = [name.stringifyOriginalName()
                                   for name in layer.getOutputTensorNames()]
            if mbox_name in output_tensor_names:
                mbox_detection_heads.append(layer)
                det_head_found = True
                # There should not be another layer that outputs the same tensor
                break

        assert det_head_found, "Could not find mbox tensor: {0}".format(mbox_name)

    # Create PermuteFlatten replacement layers.
    concat_input_tensors = []
    perm_flatten_layers = []
    for det_head in mbox_detection_heads:
        perm_flatten_shape = (1,
                              reduce(lambda x, y: x * y, det_head.getOutputTensors()[0].shape),
                              1,
                              1)
        mbox_perm_flatten_output = UnpopulatedTensor(perm_flatten_shape)
        mbox_perm_flatten_output.setName(MangledName(
            OriginalName(det_head.getOutputTensorNames()[0].stringifyOriginalName())))
        concat_input_tensors.append(mbox_perm_flatten_output)
        mbox_perm_flatten = PermuteFlatten("PERM_FLATTEN_" + det_head.getName().stringifyOriginalName(), [], [])
        mbox_perm_flatten.setInputTensorsAllFields(det_head.getInputTensors())
        mbox_perm_flatten.setOutputTensorsAllFields([mbox_perm_flatten_output])
        mbox_perm_flatten.reference_layout = Layouts.NHWC
        mbox_perm_flatten.permutation = (0, 1, 2, 3)
        perm_flatten_layers.append(mbox_perm_flatten)


    C_axis_concat_size = reduce(lambda x, y: x + y,
                                [t.shape[Layouts.Axes.C] for t in concat_input_tensors])
    mbox_concat_shape = (1, C_axis_concat_size, 1, 1)
    mbox_concat_output = UnpopulatedTensor(mbox_concat_shape)
    mbox_concat_output.setName(MangledName(OriginalName("mbox_conf")))
    mbox_concat = Concat("mbox_conf", [], [])
    mbox_concat.setInputTensorsAllFields(concat_input_tensors)
    mbox_concat.setOutputTensorsAllFields([mbox_concat_output])
    # This concat axis, normally, should be H because the Layout should be NHWC but
    # the implict concat  mechanism might not like it and because the concat inputs
    # should be flattend on the C axis it won't matter which axis is slected hence
    # use C axis.
    mbox_concat.loadConcatAxis(Layouts.Axes.C)

    # Create reshape layer.
    num_classes = det_out_config.postprocessing_params.num_classes
    mbox_concat_reshape_shape = (1,
                                 num_classes,
                                 1,
                                 C_axis_concat_size // num_classes)
    mbox_concat_reshape_output = UnpopulatedTensor(mbox_concat_reshape_shape)
    mbox_concat_reshape_output.setName(MangledName(OriginalName("mbox_conf_reshape")))
    mbox_concat_reshape = ReshapeHWC("mbox_conf_resahpe", [], [])
    mbox_concat_reshape.setInputTensorsAllFields([mbox_concat_output])
    mbox_concat_reshape.setOutputTensorsAllFields([mbox_concat_reshape_output])

    mbox_score_shape = mbox_concat_reshape_shape
    if (det_out_config.score_converter == tf_ssd_config_pb2.TFSSDConfig.SOFTMAX):
        mbox_score_output = UnpopulatedTensor(mbox_score_shape)
        mbox_score_output.setName(MangledName(OriginalName("mbox_conf_softmax")))
        mbox_score = Softmax("mbox_conf_softmax", [], [])
        mbox_score.setInputTensorsAllFields([mbox_concat_reshape_output])
        mbox_score.setOutputTensorsAllFields([mbox_score_output])
        mbox_score.setAxis(Layouts.Axes.C)
    elif (det_out_config.score_converter == tf_ssd_config_pb2.TFSSDConfig.SIGMOID):
        mbox_score_output = UnpopulatedTensor(mbox_score_shape)
        mbox_score_output.setName(MangledName(OriginalName("mbox_conf_sigmoid")))
        mbox_score = Sigmoid("mbox_conf_sigmoid", [], [])
        mbox_score.setInputTensorsAllFields([mbox_concat_reshape_output])
        mbox_score.setOutputTensorsAllFields([mbox_score_output])
    else:
        assert False, "Unrecognized SSD ScoreConverter."

    # Remove mbox_conf reshape layers.
    parsed_layers = [layer for layer in parsed_layers if layer not in mbox_detection_heads]

    # Add PermuteFlatten layers
    parsed_layers = parsed_layers + perm_flatten_layers

    parsed_layers.append(mbox_concat)
    parsed_layers.append(mbox_concat_reshape)
    parsed_layers.append(mbox_score)

    return (parsed_layers, mbox_score_output)

def _tf_ssd_pack_priors_and_vars(prior_boxes, variances):
    for var_i, var in enumerate(variances):
        prior_boxes = np.insert(prior_boxes, var_i * 2 + 1, var, axis=1)

    prior_boxes = np.expand_dims(np.expand_dims(prior_boxes, 0), 0)

    return prior_boxes

def gen_ssd_patch_layers(parsed_layers, mbox_loc_names, mbox_conf_names,
                         prior_boxes, ssd_config):
    """ Generates the operations required to replace the Postprocessor and
        MultipleGridAnchorGenerator blocks and the operations that are at global
        scope that connect the logical blocks.
    """

    parsed_layers, mbox_loc_tensor = _tf_ssd_gen_mbox_loc_patch(parsed_layers, mbox_loc_names)
    parsed_layers, mbox_conf_tensor = _tf_ssd_gen_mbox_conf_patch(parsed_layers, mbox_conf_names, ssd_config)

    variances = list(ssd_config.postprocessing_params.box_params.var)
    prior_boxes = _tf_ssd_pack_priors_and_vars(prior_boxes, variances)

    detection_out = TF_DetectionOutput("tf_detection_output", [], [])
    detection_out.load_parameters(ssd_config)
    detection_out.set_prior_boxes(prior_boxes)
    detection_out_output_shape = (1, 1, detection_out.getMaxDetections(), 7)
    detection_out_output_tensor = UnpopulatedTensor(detection_out_output_shape)
    detection_out_output_tensor.setName(MangledName(OriginalName("detection_out")))
    detection_out.setInputTensorsAllFields([mbox_loc_tensor, mbox_conf_tensor])
    detection_out.setOutputTensorsAllFields([detection_out_output_tensor])
    parsed_layers.append(detection_out)

    output_layer = Output('output', [], [])
    output_layer.setInputTensorsAllFields([detection_out_output_tensor])
    output_layer.setOutputTensorsAllFields([])
    output_layer.enableDetectionOutput()
    parsed_layers.append(output_layer)

    return parsed_layers
