# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.

import numpy as np
from Controllers.Parsers.Parser.Convolution2D import Deconvolution
import tensorflow as tf

from . import Helpers as tf_helpers

def load(obj, operations, feed_dict):

    def get_deconv_padding(input_shape, output_shape, kernel_shape, stride):
        """
            input_shape:  [H,W]
            output_shape: [H,W]
            kernel_shape: [kH,kW]
            stride:       [sH,sW]
        """
        pady = 0
        padx = 0
        # 2Xpad = stride X (input - 1) + kernel - out
        pady = stride[0] * (input_shape[0] - 1) + kernel_shape[0] - output_shape[0]
        padx = stride[1] * (input_shape[1] - 1) + kernel_shape[1] - output_shape[1]
        if (pady % 2 == 1):
            pady = -1
        else:
            pady = pady / 2
        if (padx % 2 == 1):
            padx = -1
        else:
            padx = padx / 2

        return int(pady), int(pady), int(padx), int(padx)

    # operation only supports one input Tensor and output Tensor
    assert(len(obj.inputs) == 3) # input and weights
    assert(len(obj.outputs) == 1)

    input_tensor = obj.inputs[2]
    output_tensor = obj.outputs[0]

    shape = tf_helpers.findTensorValue(obj.inputs[0].name, operations)

    inputTensorShape = tf_helpers.fillTensorDimensions(tf.shape(input_tensor).eval(feed_dict), convertToNCHW=True)
    outputTensorShape = tf_helpers.fillTensorDimensions(tf.shape(output_tensor).eval(feed_dict), convertToNCHW=True)

    x = Deconvolution(obj.name, [input_tensor.name], [output_tensor.name])
    num_groups = 1
    x.setNumGroups(num_groups)

    # NCHW
    outputChannels = outputTensorShape[1]
    x.loadOutputChannels(outputChannels)

    # ensure dilation is equal - do we support otherwise?
    try:
        # NHWC
        dilations = obj.get_attr('dilations')
        if len(set(dilations[1:3])) != 1 or dilations[1] != 1:
            raise ValueError('Unsupported Deconvolution with dilation != 1')

        x.loadDilation(dilations[1])
    except ValueError:
        # assume that dilation is default: 1
        x.loadDilation(1)

    # assume second input kernel
    kernel = obj.inputs[1]

    # The kernel is stored as HWCO
    kernelSize = tf_helpers.fillTensorDimensions(kernel.shape, False)[0:2]
    x.loadKernelSize(kernelSize[0], kernelSize[1])

    # NHWC
    stride = obj.get_attr('strides')[1:3]
    x.loadStrideSize(stride[0], stride[1])

    # inputTensor is in layout NCHW
    padding = get_deconv_padding(inputTensorShape[2:], outputTensorShape[2:],
                kernelSize, stride)
    x.loadPadding(*padding)
    x.setPadStyle(obj.get_attr('padding').decode('ascii'))

    x.setBiasEnabled(False)

    weights = tf_helpers.findTensorValue(kernel.name, operations)
    # TensorFlow's shape of Conv2DBackpropInput weights is
    #   (fH, fW, in_ch, out_ch)
    # Need to reshape to (in_ch, out_ch, fH, fW), as requested by Parser.Deconvolution
    # Taps need to be roated in the HW plane because Conv2DBackpropInput
    # is implemented via the convolution backward pass, which does an 180deg rotation.
    weights = np.transpose(weights, (2, 3, 0, 1))
    weights = weights[:,:,::-1,::-1]

    x.loadTrainedParameters(weights=weights)

    return [x]
