# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.

import PIL
from PIL import Image
import skimage
import skimage.io
import skimage.transform
import numpy as np
import re
from Models.EnumDeclarations import SeedData
import Models.Layouts as Layouts
from Controllers.EnumController import throw_error, ErrorTable, compiler_assert

RAND_HI = 1
RAND_LO = -1

def preprocess(img, args, new_size):
    """
        Prepare an input for processing
    """

    if img is None:
        print("Warning: Randomly initialized input")
    elif args.image == "Debug":
        return SeedData.random
    elif args.image == "Debug_ones":
        return SeedData.all_ones
    elif args.image == "Debug_zeros":
        return SeedData.all_zeros
    elif args.image == "Debug_int":
        return SeedData.random_int
    else:
        # Should be an image path.
        data = input_to_npy(img, new_size)
        print(data.shape)
        if (new_size is None):
            #images are HWC and we want the shape as NCHW
            data_size = (1, data.shape[2], data.shape[0], data.shape[1])
            data = standardize(data, data_size)
        else:
            data = standardize(data, new_size)
        print(data.shape)
        data = scale(data, args.raw_scale)
        print(data.shape)
        data = channel_swap(data, args.channel_swap)
        print(data.shape)
        data = mean(data, args.mean)
        print(data.shape)
        return data


def input_to_npy(path, new_size):
    """
        Convert from the input format to a numpy array.

        Currently Handles (in varying levels):
        - Images (PNG, JPG, BMP, GIF)
        - Numpy Files (saved with .save, rather than .tofile)   # TODO: Do tofile
        - Mat files, from MatLab.
    """

    # Image Parsing
    if path.split(".")[-1].lower() in ["png", "jpeg", "jpg", "bmp", "gif"]:
        greyscale = True if (new_size is not None and new_size[2] == 1) else False
        data = skimage.io.imread(
                path, as_grey=greyscale)

    # Numpy Array
    elif path.split(".")[-1] in ["npy"]:
        im = np.load(path)

        if (len(im.shape) == 2):
            if(new_size is not None and (im.shape[0] != new_size[2] or im.shape[1] != new_size[3])):
                throw_error(ErrorTable.InvalidInputFile)
        elif (len(im.shape) == 3):
            if(new_size is not None and (im.shape[0] != new_size[2] or im.shape[1] != new_size[3])):
                throw_error(ErrorTable.InvalidInputFile)
        else: #im.shape is not valid
            throw_error(ErrorTable.InvalidInputFile)
        data = np.asarray(im)

    # MAT File
    elif path.split(".")[-1] in ["mat"]:
        print("Filetype not officially supported use at your own peril: MAT File")
        import scipy.io
        im = scipy.io.loadmat(path)
        data = np.asarray(im)

    else:
        print("Unsupported Data File:", path)
        throw_error(ErrorTable.InputFileUnsupported)

    return data


def standardize(data, new_size):
    """
        Make sure that all inputs are treated the same.
    """
    # Add axis for greyscale images (size 1)
    if (len(data.shape) == 2):
        data = data[:, :, np.newaxis]

    # Transposed and reshaped. TODO: Make more obvious what is going on here.
    data = skimage.transform.resize(data, new_size[2:])
    data = np.transpose(data, (2, 0, 1))
    data = np.reshape(data, (1, data.shape[0], data.shape[1], data.shape[2]))
    return data


def scale(data, raw_scale=1):
    return data * raw_scale


def channel_swap(data, channel_swap=None):
    if channel_swap is not None and len(data.shape) == 4 and data.shape[1] == len(channel_swap):
        data[0] = data[0][np.argsort(channel_swap), :, :]
    return data


def mean(data, mean=None):
    if mean is not None:
        #  Try loading mean from .npy file
        if re.search('[a-zA-Z]+', mean):
            try:
                mean = np.load(mean)
            except:
                throw_error(ErrorTable.InvalidNpyFile, mean)

            mean = mean.mean(1).mean(1)
            mean_arr = np.zeros(data.shape[1:])

            for x in range(mean.shape[0]):
                mean_arr[x].fill(mean[x])

            data[0] -= mean_arr

        # Else, try loading mean as tuple
        elif re.search('[,]+', mean):
            try:
                (R,G,B) = mean.split(',')
            except:
                throw_error(ErrorTable.InvalidTuple, mean)

            mean = np.asarray([float(R), float(G), float(B)])
            mean_arr = np.zeros(data.shape[1:])

            for x in range(mean.shape[0]):
                mean_arr[x].fill(mean[x])

            data[0] -= mean_arr

        # Else, load mean as single number
        elif re.search(r'\d+', mean):
            try:
                data = data - float(mean)
            except:
                throw_error(ErrorTable.InvalidMean, mean)

        # Else, invalid mean input
        else:
            throw_error(ErrorTable.InvalidMean, mean)

    return data

def load_input_from_file(path):
    try:
        data = skimage.io.imread(path)
    except:
        print("Unsupported Data File:", path)
        throw_error(ErrorTable.InputFileUnsupported)
    # for grayscale images, skimage.io.imread returns an ndarray with
    # shape [M,N]; need to extend this to 3 dims (channels=1)
    if len(data.shape) == 2:
        data = np.expand_dims(data, axis=2)

    return data

def load_mean(mean):
    if mean is not None:
        try:
            return [float(m) for m in mean.split(',')]
        except:
            throw_error(ErrorTable.InvalidTuple,
                        "Invalid mean argument:{0}".format(mean))
    else:
        return None

def load_scale(scale):
    if scale is not None:
        try:
            return [float(s) for s in scale.split(',')]
        except:
            throw_error(ErrorTable.InvalidTuple,
                        "Invalid scale argument:{0}".format(scale))
    else:
        return None

def load_channel_swap(channel_swap):
    if channel_swap is not None:
        try:
            return [int(c) for c in channel_swap.split(',')]
        except:
            throw_error(ErrorTable.InvalidTuple,
                        "Invalid channel-swap argument:{0}".format(channel_swap))
    else:
        return None

def preprocess_image(image, mean, scale, channel_swap):
    num_channels = image.shape[2]
    if scale is not None:
     #   support scalar as well as per-channel scaling
     #   if len(scale) != num_channels:
     #     throw_error(ErrorTable.InvalidTuple,
     #                   "Number of scale values = {0} != number of input channels {1}".format(
     #                   len(scale), num_channels))
        scale = np.array(scale)
    else:
        scale = np.ones(num_channels)

    if mean is not None:
    #    support scalar as well as per-channel means
    #    if len(mean) != num_channels:
    #        throw_error(ErrorTable.InvalidTuple,
    #                    "Number of mean values = {0} != number of input channels {1}".format(
    #                    len(mean), num_channels))
        mean = np.array(mean)
    else:
        mean = np.zeros(num_channels)

    # Bring image into range [0, 1]
    image = image * (1.0/255.0)

    # Apply scale and shift(mean).
    image = image * scale - mean

    if channel_swap is not None:
        if len(channel_swap) != num_channels:
            throw_error(ErrorTable.InvalidTuple,
                        "Number of channel swap values = {0} != number of input channels {1}".format(
                        len(channel_swap), num_channels))
        image = image[:, :, list(channel_swap)]

    return image

def match_network_shape(image, net_input_shape):
    """ The images are read in (N)HWC layout. Check that the input image matches
        the network input shape and epand dimensions to match.

        Arguments:
            image: input numpy array to be processed which must have the HWC layout.
            net_input_shape: Input shape of the network that must be matched by the
                             shape of the image. The shape is given as an iterable
                             with the following order: N, C, H, W

        Returns:
            output: numpy array that has NHWC layout and shape = net_input_shape
    """

    image_shape = image.shape
    assert len(image_shape) >= 2, "Invalid input image shape."

    # Test if the input image shape matches the network input shape.
    assert image_shape[0] == net_input_shape[Layouts.Axes.H], \
                       "Input image height = {0} must match network input height {1}".format(
                       image_shape[0], net_input_shape[Layouts.Axes.H])

    assert image_shape[1] == net_input_shape[Layouts.Axes.W], \
                       "Input image width = {0} must match network input width {1}".format(
                       image_shape[1], net_input_shape[Layouts.Axes.W])

    if len(image_shape) == 3:
        assert image_shape[2] == net_input_shape[Layouts.Axes.C], \
                           "Input image channels = {0} must match network input channels {1}".format(
                           image_shape[2], net_input_shape[Layouts.Axes.C])
    else:
        assert net_input_shape[Layouts.Axes.C] == 1, \
                           "Input image channels = {0} must match network input channels {1}".format(
                           1, net_input_shape[Layouts.Axes.C])

    while len(image.shape) < len(net_input_shape):
        image = np.expand_dims(image, 0)

    return image

def resize(image, height, width):
    image = skimage.transform.resize(image, (height, width), preserve_range=True)
    return image

def gen_random_image(rand_type, seed, shape):
    if seed is not None:
        np.random.seed(seed)

    if rand_type is SeedData.all_ones:
        data = np.ones(shape)
    elif rand_type is SeedData.all_zeros:
        data = np.zeros(shape)
    elif rand_type is SeedData.random:
        data = np.random.uniform(RAND_LO, RAND_HI, shape)
    elif rand_type is SeedData.random_int:
        data = np.random.randint(RAND_LO, RAND_HI, shape)

    return data

