# Copyright 2018 Intel Corporation.
# The source code, information and material ("Material") contained herein is
# owned by Intel Corporation or its suppliers or licensors, and title to such
# Material remains with Intel Corporation or its suppliers or licensors.
# The Material contains proprietary information of Intel or its suppliers and
# licensors. The Material is protected by worldwide copyright laws and treaty
# provisions.
# No part of the Material may be used, copied, reproduced, modified, published,
# uploaded, posted, transmitted, distributed or disclosed in any way without
# Intel's prior express written permission. No license under any patent,
# copyright or other intellectual property rights in the Material is granted to
# or conferred upon you, either expressly, by implication, inducement, estoppel
# or otherwise.
# Any license under such intellectual property rights must be express and
# approved by Intel in writing.

import os
import sys

import numpy as np

from Models.Network import *
from Models.NetworkStage import *
from Models.MyriadParam import *
from Models.EnumDeclarations import *
from Controllers.Parsers.Caffe import CaffeParser
from Controllers.Parsers.TensorFlow import TensorFlowParser
from Controllers.PingPong import getManualHwSchedule
from Views.IRVisualize import drawIR, drawGraphFromLayers, drawGraph
from Controllers.CPPWrapper import CPP_INTERCEPT
import Controllers.TensorFormatSelection as tensorFormatSelection
import Models.Layouts as Layouts
import Controllers.Optimizer as optimizer
import Controllers.Parsers.Phases as phases

import Controllers.Globals as GLOBALS

def load_myriad_config(no_shaves):
    return MyriadParam(0, no_shaves - 1)

def load_network(arguments, parser, myriad_conf, debug=False):
    """
    Loads the information contained in a network description file into our internal representation.
    It does not calculate buffer pointers for myriad. This is done in a seperate step.
    :param weights:
    :param path: absolute path to network description file
    :param parser: enum indicating which parser should be used.
    :return: Network with attributes populated, bar inputs and outputs.
    """

    if hasattr(arguments, "nce_slices") and arguments.nce_slices is not None:
        GLOBALS.NCE_CMX_SLICES = arguments.nce_slices

    # For consistency, do this always
    GLOBALS.NCE_CMX_SIZE = GLOBALS.NCE_CMX_SLICES * GLOBALS.CMX_SLICE_SIZE

    if not(arguments.old_parser) or arguments.cpp:

        if parser == Parser.Caffe:
            p = CaffeParser()
        elif parser == Parser.TensorFlow:
            p = TensorFlowParser()
        else:
            throw_error(ErrorTable.ParserNotSupported, parser.name)

        # Load From Framework
        p.loadNetworkObjects(arguments.net_description, arguments.net_weights)

        if isinstance(p, TensorFlowParser) and arguments.tf_ssd_config is not None:
            p.load_ssd_config(arguments.tf_ssd_config)

        parsedLayers = p.parse(arguments)

        if parser == Parser.TensorFlow and arguments.tf_ssd_config is not None:
            mbox_loc_names, mbox_conf_names = p.get_mbox_loc_conf_tensor_names()
            prior_boxes = p.get_prior_boxes(arguments)
            parsedLayers = phases.gen_ssd_patch_layers(parsedLayers,
                                                       mbox_loc_names,
                                                       mbox_conf_names,
                                                       prior_boxes,
                                                       p.get_ssd_config())

        input_data, expected_result = p.calculateReference(arguments)

        if(arguments.cpp):
            CPP_INTERCEPT(parsedLayers)
            quit()

        if arguments.ma2480 and parser == Parser.TensorFlow:
            if arguments.force_TF_symmetric_pad:
                parsedLayers = phases.forceSymmetricalPadding(parsedLayers)
            else:
                parsedLayers = phases.transformForHwImplementation(parsedLayers)

        parsedLayers = phases.overrideSoftwareSetting(parsedLayers, getManualHwSchedule())

        # Apply Transforms
        if parser != Parser.TensorFlow:
            parsedLayers = phases.cropNetwork(parsedLayers, arguments)
        parsedLayers = phases.breakUpInnerLRN(parsedLayers)

        if parser == Parser.Caffe:
            parsedLayers = phases.caffeCropAdjustment(parsedLayers)

        parsedLayers = phases.selectInputOutputLayout(parsedLayers, arguments.input_layout, arguments.output_layout)
        parsedLayers = optimizer.postParsingOptimizations(parsedLayers)

        parsedLayers = phases.splitGroupDeconvolution(parsedLayers)

        bypass_opt = arguments.scheduler is not None
        bypass_opt = bypass_opt if parser == Parser.Caffe else False

        scheduler = NCE_Scheduler(manual_schedule=getManualHwSchedule())
        parsedLayers = optimizer.selectImplementations(
                parsedLayers, scheduler, arguments.ma2480, bypass_opt=bypass_opt)
        if arguments.accuracy_table != {} and not bypass_opt:
            parsedLayers = phases.adaptHwWeights(
                p, parsedLayers, input_range=(np.min(input_data), np.max(input_data)))

        parsedLayers = phases.splitGroupConvolutionForShaves(parsedLayers)

        parsedLayers = phases.Slice_with_port_crop_layers(parsedLayers)

        parsedLayers = optimizer.postImplementationOptimizations(parsedLayers)

        parsedLayers = tensorFormatSelection.select_tensors_format(parsedLayers)

        if bypass_opt:
            phases.forceTensorLayout(parsedLayers)

        parsedLayers = phases.implicitConcatBinding(parsedLayers)

        parsedLayers = phases.squashInPlaceLayers(parsedLayers)

        parsedLayers = phases.placeCroppedTensor(parsedLayers)

        parsedLayers = phases.solveRefineDetSpeicificDetectionOutput(parsedLayers)

        net, graphFile, finalLayers = phases.serializeNewFmt(parsedLayers, arguments, myriad_conf, input_data)

        network_dict = {
            'network': net,
            'expected': expected_result,
            'expected_shape': expected_result.shape,
            'expected_layout': None,
            'graph': graphFile,
            'final_layers': finalLayers  # full final layer list including implicit layers to facilitate profile graph creation
        }
        return network_dict

    else:

        if parser == Parser.Debug:
            throw_error(ErrorTable.ParserNotSupported)
        elif parser == Parser.TensorFlow:
            from Controllers.TensorFlowParser import parse_tensor
            parse_ret = parse_tensor(arguments, myriad_conf)
        elif parser == Parser.Caffe:
            from Controllers.CaffeParser import parse_caffe
            parse_ret = parse_caffe(arguments, myriad_conf)
        else:
            throw_error(ErrorTable.ParserNotSupported, parser.name)

        network = parse_ret['network']

        network.finalize()
        network.optimize()
        if arguments.ma2480:
            network.convert_for_hardware()

        return parse_ret
